import sys
from queue import Queue
from struct import unpack
from threading import Thread
from time import sleep, time

import numpy as np
from PyQt5.QtGui import QPixmap, QImage
from PyQt5.QtWidgets import (QWidget, QPushButton,
                             QApplication, QLabel)
from serial import Serial
from serial.tools.list_ports import comports


class UbRGBColorSensor(QWidget):
    def __init__(self):
        super().__init__()
                
        # Serial port
        self._port_device = None
        self._ser = None
        self.connected = False
        
        # Run status
        self.running = False
        
        # Initialize queue
        self.q = Queue()
        self._sentinel = object()
        
        # Threads
        self.tcmd = None  # Commands
        self.tp = None  # Producer
        self.tc = None  # Consumer
        
        # Image
        self.w = 256
        self.h = 256
        self.image_data = None
        
        # Labels
        self.image_label = None
        self.status_label = None
        self.rgb_label = None
        self.data_label = None
        
        # TCS34725 Default
        self.rgbc = None
        self.again = 2
        self.atime = 192
        self.rgbc_count = 65535

        # Initialize UI
        self.init_ui()
        self.connect()
        self.start_cmd()
        if self.connected:
            self.start_pc()

    def init_ui(self):
        # Label: status
        status_label = QLabel(self)
        status_label.move(20, 10)
        self.status_label = status_label

        # Image: RGB Colors
        self.image_data = np.zeros((self.h, self.w, 3), dtype=np.uint8)
        image_label = QLabel(self)
        image_label.move(0, 35)
        self.image_label = image_label
        self.set_image((255, 255, 255))

        # Values: Data
        data_label = QLabel(self)
        data_label.move(10, 310)
        data_label.setText("Data\nR:    \nG:    \nB:    \nC:")
        self.data_label = data_label

        # Values: RGB
        rgb_label = QLabel(self)
        rgb_label.move(75, 310)
        rgb_label.setText("RGB\nR: 255\nG: 255\nB: 255")
        self.rgb_label = rgb_label

        # App Window
        self.setGeometry(50, 100, 256, 410)
        self.setWindowTitle('micro:bit  RGB Color Sensor')
        self.show()

    def cmd(self):
        while True:               
            line = sys.stdin.readline()
            if line[0:7] == 'connect':
                self.connect()
                self.start_pc()
            elif line[0:7] == 'capture':
                print(self.rgbc)
                sys.stdout.flush()
            elif line[0:5] == 'again':
                line_ = line.rstrip().split()
                try:
                    again = int(line_[1])
                except:
                    pass
                else:
                    if 0 <= again <= 3:
                        ub_cmd = 'set_again(%d)' % again
                        self._ser.write(ub_cmd.encode('utf-8'))
                        self.again = again
            elif line[0:5] == 'atime':
                line_ = line.rstrip().split()
                try:
                    atime = int(line_[1])
                except:
                    pass
                else:
                    if 0 <= atime <= 255:
                        ub_cmd = 'set_atime(%d)' % atime
                        self._ser.write(ub_cmd.encode('utf-8'))  
                        self.max_rgbc_count(atime)
                        self.atime = atime
                        

                
            
    def start_cmd(self):
        tcmd = Thread(target=self.cmd)
        tcmd.daemon = True
        tcmd.start()
        self.tcmd = tcmd
        
    def start_pc(self):
        if self.connected:
            self.running = True
            tp = Thread(target=self.producer, args=(self.q,))
            tc = Thread(target=self.consumer, args=(self.q,))
            tp.daemon = True
            tc.daemon = True
            tp.start()
            tc.start()
            self.tp = tp
            self.tc = tc  
        else:
            self.set_status("micro:bit not connected")

    def find_microbit(self):
        microbit_pid = 516
        microbit_vid = 3368
        """
        Finds the port for the first micro:bit found connected to the computer.
        """
        for port in comports():
            if port.vid == microbit_vid and port.pid == microbit_pid:
                self._port_device = port.device

    def connect(self):
        """
        Initialize a pySerial Serial object to talk to the microbit
        """
        baudrate = 115200
        parity = 'N'
        self.find_microbit()
        if self._port_device is None:
            self.set_status("No micro:bit found")
        else:
            try:
                s = Serial(self._port_device, baudrate, parity=parity)
            except:
                self.set_status("Could not open port %s" % self._port_device)
            else:
                self.connected = True
                self.set_status("micro:bit connected")        
                self._ser = s

    def terminate(self):
        self.running = False

    def producer(self, out_q):
        """ 
        Data from micro:bit to queue
        """
        self._ser.reset_input_buffer()
        while self.running:
            buf = bytearray()
            while not buf.endswith(b'\r\n'):
                buf.extend(self._ser.read_all())
                sleep(0.001)
                if not self.running:
                    break
            if self.running:
                buflist = buf.split(b'\r\n')
                for i in buflist[:-1]:
                    if len(i) == 8:
                        out_q.put(unpack("<HHHH", i))
            sleep(0.001)
        out_q.put(self._sentinel)

    def consumer(self, in_q):
        """ 
        Data from queue processed and displayed in UI 
        """
        t_prev = time()
        while True:
            data = in_q.get()
            if data is self._sentinel:
                break
            r, g, b, c = data
            max_count = self.rgbc_count
            rgb = [int(r / max_count * 255), int(g / max_count * 255), int(b / max_count * 255)]
            t_now = time()
            dt = t_now - t_prev
            if dt > 0.05:
                self.set_data(data)
                self.set_rgb(rgb)
                self.set_image(rgb)
                t_prev = t_now
            self.rgbc = data
            sleep(0.0001)
            
    def max_rgbc_count(self, atime):
        max_count = (256 - atime) * 1024
        if max_count <= 65535:
            self.rgbc_count = max_count
        else:
            self.rgbc_count = 65535
            
    def set_image(self, rgb):
        """ 
        TBC
        """
        r, g, b = rgb
        self.image_data[:, :] = rgb
        h = 76
        vr = 256 - r / 256 * h
        vg = 256 - g / 256 * h
        vb = 256 - b / 256 * h
        self.image_data[179:180, :] = [200, 200, 200]
        self.image_data[int(vr):256, 7:83] = [r, 0, 0]
        self.image_data[int(vg):256, 90:166] = [0, g, 0]
        self.image_data[int(vb):256, 172:248] = [0, 0, b]
        img = QImage(self.image_data.tobytes(), self.w, self.h, QImage.Format_RGB888)
        self.image_label.setPixmap(QPixmap.fromImage(img))

    def set_status(self, status):
        self.status_label.setText(status)
        self.status_label.adjustSize()

    def set_data(self, data):
        """ 
        TBC
        """
        r, g, b, c = data
        data_str = "Data\nR: %d\nG: %d\nB: %d\nC: %d" % (r, g, b, c)
        self.data_label.setText(data_str)
        self.data_label.adjustSize()

    def set_rgb(self, rgb):
        """ 
        TBC
        """
        r, g, b = rgb
        rgb_str = "RGB\nR: %d\nG: %d\nB: %d" % (r, g, b)
        self.rgb_label.setText(rgb_str)
        self.rgb_label.adjustSize()


if __name__ == '__main__':
    app = QApplication(sys.argv)
    ubrgb = UbRGBColorSensor()
    sys.exit(app.exec_())
